//////////////////////////////////////////////////////////////////////////////
// main.cpp                                                                 //
//////////////////////////////////////////////////////////////////////////////
#include <windows.h>
#include <stdio.h>
#include <conio.h>

//////////////////////////////////////////////////////////////////////////////
// defines                                                                  //
//////////////////////////////////////////////////////////////////////////////
#define PORT	0x378
#define SDO		1
#define SCL		2
#define VCC		(0xFF - SDO - SCL)
#define SDI		(_inp(PORT+1) & 0x20 ? 1 : 0)	// pin 12

#define Read(addr) ( ReadBlock(addr, 1), Transfer() )
#define Write(addr,data) ( WriteBlock(addr, 1), Transfer(data) )

//////////////////////////////////////////////////////////////////////////////
// Transfer                                                                 //
//////////////////////////////////////////////////////////////////////////////
unsigned char Transfer(unsigned char out = 0xFF)
{
	unsigned char in=0;
	for (int bit=7; bit>=0; bit--)
	{
		int n = out&0x80 ? VCC+SDO : VCC;
		out <<= 1;
		_outp(PORT, n);
		in = in<<1 | SDI;
		_outp(PORT, n | SCL);
		_outp(PORT, n);
		//for (volatile long d=1000; d; d--);
	}
	for (volatile long d=10000; d; d--);
	return in;
}

//////////////////////////////////////////////////////////////////////////////
// Writelock                                                                //
//////////////////////////////////////////////////////////////////////////////
void WriteBlock(unsigned int address, unsigned long size)
{
	Transfer('w');
	Transfer(address>>8);
	Transfer(address);
	Transfer((unsigned char)(size>>8));
	Transfer((unsigned char)size);
}

//////////////////////////////////////////////////////////////////////////////
// ReadBlock                                                                //
//////////////////////////////////////////////////////////////////////////////
void ReadBlock(unsigned int address, unsigned long size)
{
	Transfer('r');
	Transfer(address>>8);
	Transfer(address);
	Transfer((unsigned char)(size>>8));
	Transfer((unsigned char)size);
}

//////////////////////////////////////////////////////////////////////////////
// Jump                                                                     //
//////////////////////////////////////////////////////////////////////////////
void Jump(unsigned int address)
{
	Transfer('j');
	Transfer(address>>8);
	Transfer(address);
}

//////////////////////////////////////////////////////////////////////////////
// Call                                                                     //
//////////////////////////////////////////////////////////////////////////////
void Call(unsigned int address)
{
	Transfer('c');
	Transfer(address>>8);
	Transfer(address);
}

//////////////////////////////////////////////////////////////////////////////
// EnableInterrupts                                                         //
//////////////////////////////////////////////////////////////////////////////
void EnableInterrupts()
{
	Transfer('e');
}

//////////////////////////////////////////////////////////////////////////////
// DisableInterrupts                                                        //
//////////////////////////////////////////////////////////////////////////////
void DisableInterrupts()
{
	Transfer('d');
}

//////////////////////////////////////////////////////////////////////////////
// Stop                                                                     //
//////////////////////////////////////////////////////////////////////////////
void Stop()
{
	DisableInterrupts();
	//Write(0xFFFF, 0x18);	// joypad, sio
	//EnableInterrupts();
	Transfer('s');
}

//////////////////////////////////////////////////////////////////////////////
// main                                                                     //
//////////////////////////////////////////////////////////////////////////////
void main()
{
	// open & close GIVEIO device for direct port access
	HANDLE h = CreateFile("\\\\.\\giveio", 0, 0, NULL, OPEN_ALWAYS, 0, NULL);
	if (h == INVALID_HANDLE_VALUE) { printf("GiveIO error\n"); return; }
	CloseHandle(h);

	printf("Serial test program\n");


	while (1)
	{
		char c = getch();
		if (c == 27) break;

		if (c == '0') Call(0x0100);
		if (c == '1') Call(0x42BA);
		if (c == '2') Call(0x0939);
		if (c == '3') Call(0x0B17);
		if (c == '4') Call(0x0406);
		if (c == '5') Call(0x02FC);
		if (c == '6') Call(0x0956);
		if (c == '7') Call(0x5F7A);
		if (c == '8') Call(0x303B);
		if (c == '9') Call(0x0B17);
		if (c == 'A') Call(0x302F);	// main
		if (c == 'B') Call(0x0939);


		if (c == ' ')	// 
		{
			Write(0x0000, 0x00);	// ram enable (0A)
			//Write(0x2000, 0x01);	// rom bank
			Write(0x4000, 0x10);
			//Write(0x6000, 0x00);	// mem model


			// wait
			Write(0xA000, 0x03);
			while (Read(0xA000) & 1);


			Write(0xA001, 1 + 0xA0);
			Write(0xA002, 0x10);
			Write(0xA003, 0);
			Write(0xA005, 1 + 0x80);



			/*printf("%02X\n", Read(0xA000));
			Sleep(500);
			printf("%02X\n", Read(0xA000));*/

			int weird[16] = { 0x06, 0x24, 0x1E, 0x0C, 0x15, 0x33, 0x2D, 0x1B, 0x12, 0x30, 0x2A, 0x18, 0x09, 0x27, 0x21, 0x0F };
			int i, j;

			j = 0x8900;
			for (i=0; i<16; i++)
			{
				Write(0xA000+weird[i], j>>8);
				j+=0x90;
			}
			j = 0x9200;
			for (i=0; i<16; i++)
			{
				Write(0xA001+weird[i], j>>8);
				j+=0x100;
			}
			j = 0xA200;
			for (i=0; i<16; i++)
			{
				Write(0xA002+weird[i], j>>8);
				j+=0x300;
			}


			// wait
			Write(0xA000, 0xFF);
			while (Read(0xA000) & 1);


			/*Write(0xA001, 1 + 0xA0);
			Write(0xA002, 0x10);
			Write(0xA003, 0);
			Write(0xA005, 1 + 0x80);*/




/*
A001 = n | A0
A002 = 10
A003 = 0
A005 = n | 80


A100-AADA data?


*/


			//5091 -> read two bytes from Axxx


			//h=73
			//l=(Axxx) --.
			//           |
			//h=00       |
			//l=(hl)  <-/



			//A320,321
			//336,337
			//34C,34D
			//352,353
			//368
			//423


			//A000: bit 0-2
			
			/*Write(0xA000, 0xFF);
			Write(0xA001, 0xFF);
			Write(0xA002, 0xFF);
			Write(0xA003, 0xFF);*/

			printf("{");
			for (i=0; i<512; i++)
			{
				printf("%02X,", Read(0xA000+i));
			}
			printf("}\n");

			/*while (!kbhit())
			{
				Write(rand()&0x7FFF, rand());
				printf("%02X\n", Read(0x0000));
			}*/
		};

		if (c == 't')	// get title
		{
			// test HRAM
			Write(0xFF80, 0x55);
			Write(0xFF81, 0xAA);
			if (Read(0xFF80) != 0x55) printf("Not responding!\n");

			// print title
			int i=0x10;
			ReadBlock(0x0134, i);
			printf("Title: \"");
			while (i--) printf("%c", Transfer(rand())); 
			printf("\"\n");
		}

		if (c == 's')	// stop/switch
		{
			Stop();
			printf("Stopped... Press button on GameBoy to resume...\n");
		}
	}
}
